@@ -7,6 +7,8 @@ from typing import Tuple
 
 # backrooms
 from backrooms.translator import StringHandler
+from . import variables
+from . import heap
 
 
 def get_builtins() -> Tuple[StringHandler, ...]:
@@ -14,5 +16,5 @@ def get_builtins() -> Tuple[StringHandler, ...]:
     info: Gets Backrooms builtins modules.
     :return: Tuple[StringHandler, ...]
     """
-    # TODO write me
-    return ()
+    return (variables.get_handler(),
+            heap.get_handler())
