@@ -104,7 +104,11 @@ class RuleModule(Rule):
                  rules: Tuple):
         super(RuleModule, self).__init__(start_character, work_space)
         rules_obj = [rule(work_space) for rule in rules]
-        self._rules: Dict[str, Rule] = {rule.get_start_character(): rule for rule in rules_obj}
+        self._rules: Dict[str, Rule] = {}
+        for rule in rules_obj:
+            if rule.get_start_character() in self._rules:
+                raise RuleError.start_character_collection(rule.get_start_character())
+            self._rules[rule.get_start_character()] = rule
 
     def __call__(self,
                  portal: 'backrooms.portal.Portal',
@@ -160,7 +164,7 @@ class Halt(Rule):
 class Output(Rule):
     def __init__(self,
                  work_space: WorkSpace):
-        super(Output, self).__init__("o", work_space)
+        super(Output, self).__init__("e", work_space)
 
     def __call__(self,
                  portal: 'backrooms.portal.Portal',
@@ -188,7 +192,7 @@ class Output(Rule):
 class Input(Rule):
     def __init__(self,
                  work_space: WorkSpace):
-        super(Input, self).__init__("i", work_space)
+        super(Input, self).__init__("c", work_space)
 
     def __call__(self,
                  portal: 'backrooms.portal.Portal',
@@ -367,16 +371,22 @@ class Down(Shifter):
         super(Down, self).__init__(0, -1, 0, "v", work_space)
 
 
+class DownUpper(Shifter):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(DownUpper, self).__init__(0, -1, 0, "V", work_space)
+
+
 class Upper(Shifter):
     def __init__(self,
                  work_space: WorkSpace):
-        super(Upper, self).__init__(0, 0, 1, "/", work_space)
+        super(Upper, self).__init__(0, 0, 1, "{", work_space)
 
 
 class Lower(Shifter):
     def __init__(self,
                  work_space: WorkSpace):
-        super(Lower, self).__init__(0, 0, -1, "\\", work_space)
+        super(Lower, self).__init__(0, 0, -1, "}", work_space)
 
 
 class Branch(Rule):
@@ -488,6 +498,55 @@ class Pop(Rule):
         yield
 
 
+class PopFrame(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(PopFrame, self).__init__("a", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        conscious[c.WORK_STACK].pop_frame()
+        conscious.step()
+        yield
+
+
+class Switch(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(Switch, self).__init__("z", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        item_2 = conscious[c.WORK_STACK].pop()
+        item_1 = conscious[c.WORK_STACK].pop()
+        conscious[c.WORK_STACK].push(item_2)
+        conscious[c.WORK_STACK].push(item_1)
+        conscious.step()
+        yield
+
+
 class Duplicate(Rule):
     def __init__(self,
                  work_space: WorkSpace):
@@ -873,6 +932,116 @@ class StringJoin(Rule):
         yield
 
 
+class StringEqual(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(StringEqual, self).__init__("e", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        string_2 = conscious[c.WORK_STACK].pop()
+        string_1 = conscious[c.WORK_STACK].pop()
+        if isinstance(string_1, str) and isinstance(string_2, str):
+            conscious[c.WORK_STACK].push(int(string_1 == string_2))
+        else:
+            conscious[c.WORK_STACK].push(None)
+        conscious.step()
+        yield
+
+
+class StringIn(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(StringIn, self).__init__("i", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        string_2 = conscious[c.WORK_STACK].pop()
+        string_1 = conscious[c.WORK_STACK].pop()
+        if isinstance(string_1, str) and isinstance(string_2, str):
+            conscious[c.WORK_STACK].push(int(string_1 in string_2))
+        else:
+            conscious[c.WORK_STACK].push(None)
+        conscious.step()
+        yield
+
+
+class StringUpper(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(StringUpper, self).__init__("u", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        item = conscious[c.WORK_STACK].pop()
+        if isinstance(item, str):
+            conscious[c.WORK_STACK].push(item.upper())
+        else:
+            conscious[c.WORK_STACK].push(None)
+        conscious.step()
+        yield
+
+
+class StringLower(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(StringLower, self).__init__("o", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        item = conscious[c.WORK_STACK].pop()
+        if isinstance(item, str):
+            conscious[c.WORK_STACK].push(item.lower())
+        else:
+            conscious[c.WORK_STACK].push(None)
+        conscious.step()
+        yield
+
+
 class StringModule(RuleModule):
     def __init__(self,
                  work_space: WorkSpace):
@@ -883,7 +1052,11 @@ class StringModule(RuleModule):
                                             StringAt,
                                             StringByte,
                                             StringSplit,
-                                            StringJoin))
+                                            StringJoin,
+                                            StringEqual,
+                                            StringIn,
+                                            StringUpper,
+                                            StringLower))
 
 
 class IntegerCast(Rule):
@@ -1007,10 +1180,37 @@ class IntegerByte(Rule):
         yield
 
 
+class IntegerAbsolute(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(IntegerAbsolute, self).__init__("l", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        item = conscious[c.WORK_STACK].pop()
+        if isinstance(item, int):
+            conscious[c.WORK_STACK].push(abs(item))
+        else:
+            conscious[c.WORK_STACK].push(None)
+        conscious.step()
+        yield
+
+
 class IntegerModule(RuleModule):
     def __init__(self,
                  work_space: WorkSpace):
-        super(IntegerModule, self).__init__("a",
+        super(IntegerModule, self).__init__("i",
                                             work_space,
                                             (IntegerCast,
                                              IntegerAdd,
@@ -1019,7 +1219,8 @@ class IntegerModule(RuleModule):
                                              IntegerDivide,
                                              IntegerModular,
                                              IntegerPower,
-                                             IntegerByte))
+                                             IntegerByte,
+                                             IntegerAbsolute))
 
 
 class X(Rule):
@@ -1145,13 +1346,37 @@ class ThreadJoin(Rule):
         yield
 
 
+class ThreadID(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(ThreadID, self).__init__("i", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        conscious[c.WORK_STACK].push(conscious[c.ID])
+        conscious.step()
+        yield
+
+
 class ThreadModule(RuleModule):
     def __init__(self,
                  work_space: WorkSpace):
         super(ThreadModule, self).__init__("t",
                                            work_space,
                                            (ThreadThread,
-                                            ThreadJoin))
+                                            ThreadJoin,
+                                            ThreadID))
 
 
 class HallwayCall(Rule):
@@ -1184,6 +1409,16 @@ class HallwayCall(Rule):
             conscious[c.FUNCTION_STACK].push(conscious[c.PC_V_X])
             conscious[c.FUNCTION_STACK].push(conscious[c.PC_V_Y])
             conscious[c.FUNCTION_STACK].push(conscious[c.PC_V_FLOOR])
+            conscious[c.FUNCTION_STACK].push(conscious[c.R0])
+            conscious[c.FUNCTION_STACK].push(conscious[c.R1])
+            conscious[c.FUNCTION_STACK].push(conscious[c.R2])
+            conscious[c.FUNCTION_STACK].push(conscious[c.R3])
+            conscious[c.FUNCTION_STACK].push(conscious[c.R4])
+            conscious[c.FUNCTION_STACK].push(conscious[c.R5])
+            conscious[c.FUNCTION_STACK].push(conscious[c.R6])
+            conscious[c.FUNCTION_STACK].push(conscious[c.R7])
+            conscious[c.FUNCTION_STACK].push(conscious[c.R8])
+            conscious[c.FUNCTION_STACK].push(conscious[c.R9])
             conscious[c.PC_X] = 0
             conscious[c.PC_Y] = hallway
         else:
@@ -1209,8 +1444,8 @@ class HallwayLevelCall(Rule):
         :param rule_step_visuals: List[Tuple[int, int, int]]
         :return: Generator[None, None, None]
         """
-        hallway = conscious[c.WORK_STACK].pop()
         floor = conscious[c.WORK_STACK].pop()
+        hallway = conscious[c.WORK_STACK].pop()
         if isinstance(floor, str):
             floor = rooms.get_floor_level(floor)
         if isinstance(floor, int):
@@ -1225,6 +1460,16 @@ class HallwayLevelCall(Rule):
                 conscious[c.FUNCTION_STACK].push(conscious[c.PC_V_X])
                 conscious[c.FUNCTION_STACK].push(conscious[c.PC_V_Y])
                 conscious[c.FUNCTION_STACK].push(conscious[c.PC_V_FLOOR])
+                conscious[c.FUNCTION_STACK].push(conscious[c.R0])
+                conscious[c.FUNCTION_STACK].push(conscious[c.R1])
+                conscious[c.FUNCTION_STACK].push(conscious[c.R2])
+                conscious[c.FUNCTION_STACK].push(conscious[c.R3])
+                conscious[c.FUNCTION_STACK].push(conscious[c.R4])
+                conscious[c.FUNCTION_STACK].push(conscious[c.R5])
+                conscious[c.FUNCTION_STACK].push(conscious[c.R6])
+                conscious[c.FUNCTION_STACK].push(conscious[c.R7])
+                conscious[c.FUNCTION_STACK].push(conscious[c.R8])
+                conscious[c.FUNCTION_STACK].push(conscious[c.R9])
                 conscious[c.PC_X] = 0
                 conscious[c.PC_Y] = hallway
                 conscious[c.PC_FLOOR] = floor
@@ -1253,6 +1498,16 @@ class HallwayReturn(Rule):
         :param rule_step_visuals: List[Tuple[int, int, int]]
         :return: Generator[None, None, None]
         """
+        r9 = conscious[c.FUNCTION_STACK].pop()
+        r8 = conscious[c.FUNCTION_STACK].pop()
+        r7 = conscious[c.FUNCTION_STACK].pop()
+        r6 = conscious[c.FUNCTION_STACK].pop()
+        r5 = conscious[c.FUNCTION_STACK].pop()
+        r4 = conscious[c.FUNCTION_STACK].pop()
+        r3 = conscious[c.FUNCTION_STACK].pop()
+        r2 = conscious[c.FUNCTION_STACK].pop()
+        r1 = conscious[c.FUNCTION_STACK].pop()
+        r0 = conscious[c.FUNCTION_STACK].pop()
         v_floor = conscious[c.FUNCTION_STACK].pop()
         v_y = conscious[c.FUNCTION_STACK].pop()
         v_x = conscious[c.FUNCTION_STACK].pop()
@@ -1261,6 +1516,16 @@ class HallwayReturn(Rule):
         x = conscious[c.FUNCTION_STACK].pop()
         if isinstance(v_x, int) and isinstance(v_y, int) and isinstance(v_floor, int):
             if isinstance(x, int) and isinstance(y, int) and isinstance(floor, int):
+                conscious[c.R0] = r0
+                conscious[c.R1] = r1
+                conscious[c.R2] = r2
+                conscious[c.R3] = r3
+                conscious[c.R4] = r4
+                conscious[c.R5] = r5
+                conscious[c.R6] = r6
+                conscious[c.R7] = r7
+                conscious[c.R8] = r8
+                conscious[c.R9] = r9
                 conscious[c.PC_X] = x
                 conscious[c.PC_Y] = y
                 conscious[c.PC_FLOOR] = floor
@@ -1271,6 +1536,36 @@ class HallwayReturn(Rule):
         yield
 
 
+class HallwayGetName(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(HallwayGetName, self).__init__("n", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        floor = conscious[c.WORK_STACK].pop()
+        hallway = conscious[c.WORK_STACK].pop()
+        if isinstance(hallway, int) and isinstance(floor, int):
+            hallway = rooms.find_hallway_location(hallway, floor)
+            if hallway is not None:
+                conscious[c.WORK_STACK].push(rooms.get_hallway_name(hallway, floor))
+            else:
+                conscious[c.WORK_STACK].push(None)
+        conscious.step()
+        yield
+
+
 class HallwayGetLocation(Rule):
     def __init__(self,
                  work_space: WorkSpace):
@@ -1289,7 +1584,40 @@ class HallwayGetLocation(Rule):
         :param rule_step_visuals: List[Tuple[int, int, int]]
         :return: Generator[None, None, None]
         """
-        conscious[c.WORK_STACK].push(rooms.find_hallway_location(conscious[c.PC_Y], conscious[c.PC_FLOOR]))
+        floor = conscious[c.WORK_STACK].pop()
+        hallway = conscious[c.WORK_STACK].pop()
+        if isinstance(hallway, str) and isinstance(floor, int):
+            conscious[c.WORK_STACK].push(rooms.get_hallway_location(floor, hallway))
+        elif isinstance(hallway, int) and isinstance(floor, int):
+            conscious[c.WORK_STACK].push(rooms.find_hallway_location(hallway, floor))
+        conscious.step()
+        yield
+
+
+class HallwaySet(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(HallwaySet, self).__init__("s", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        floor = conscious[c.WORK_STACK].pop()
+        hallway = conscious[c.WORK_STACK].pop()
+        hallway_name = conscious[c.WORK_STACK].pop()
+        if isinstance(floor, int) and isinstance(hallway, int)\
+                and (isinstance(hallway_name, str) or hallway_name is None):
+            rooms.set_hallway_name(hallway, floor, hallway_name)    # error handling
         conscious.step()
         yield
 
@@ -1312,7 +1640,63 @@ class HallwayRemove(Rule):
         :param rule_step_visuals: List[Tuple[int, int, int]]
         :return: Generator[None, None, None]
         """
-        # TODO write
+        hallway = conscious[c.WORK_STACK].pop()
+        floor = conscious[c.WORK_STACK].pop()
+        if isinstance(hallway, int) and isinstance(floor, int):
+            rooms.remove_hallway(hallway, floor)
+        conscious.step()
+        yield
+
+
+class HallwayPast(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(HallwayPast, self).__init__("p", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        hallway = conscious[c.WORK_STACK].pop()
+        floor = conscious[c.WORK_STACK].pop()
+        if isinstance(hallway, int) and isinstance(floor, int):
+            conscious[c.WORK_STACK].push(rooms.get_past_hallway_location(hallway, floor))
+        conscious.step()
+        yield
+
+
+class HallwayNext(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(HallwayNext, self).__init__("e", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        hallway = conscious[c.WORK_STACK].pop()
+        floor = conscious[c.WORK_STACK].pop()
+        if isinstance(hallway, int) and isinstance(floor, int):
+            conscious[c.WORK_STACK].push(rooms.get_next_hallway_location(hallway, floor))
+        conscious.step()
         yield
 
 
@@ -1324,7 +1708,268 @@ class HallwayModule(RuleModule):
                                             (HallwayCall,
                                              HallwayLevelCall,
                                              HallwayReturn,
-                                             HallwayGetLocation))
+                                             HallwayGetLocation,
+                                             HallwayGetName,
+                                             HallwaySet,
+                                             HallwayRemove,
+                                             HallwayPast,
+                                             HallwayNext))
+
+
+class HallwayGetFloorName(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(HallwayGetFloorName, self).__init__("n", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        floor = conscious[c.WORK_STACK].pop()
+        if isinstance(floor, int):
+            conscious[c.WORK_STACK].push(rooms.get_floor_name(floor))
+        conscious.step()
+        yield
+
+
+class HallwayGetFloorLevel(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(HallwayGetFloorLevel, self).__init__("l", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        floor_name = conscious[c.WORK_STACK].pop()
+        if isinstance(floor_name, str):
+            conscious[c.WORK_STACK].push(rooms.get_floor_level(floor_name))
+        conscious.step()
+        yield
+
+
+class HallwaySetFloorLevel(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(HallwaySetFloorLevel, self).__init__("s", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        floor_name = conscious[c.WORK_STACK].pop()
+        floor_level = conscious[c.WORK_STACK].pop()
+        if isinstance(floor_level, int) and (isinstance(floor_name, str) or floor_name is None):
+            rooms.set_floor_name(floor_level, floor_name)
+        conscious.step()
+        yield
+
+
+class LevelModule(RuleModule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(LevelModule, self).__init__("l",
+                                          work_space,
+                                          (HallwayGetFloorName,
+                                           HallwayGetFloorLevel,
+                                           HallwaySetFloorLevel))
+
+
+class ReadFlip(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(ReadFlip, self).__init__("r", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        x = conscious[c.PC_X]
+        y = conscious[c.PC_Y]
+        floor = conscious[c.PC_FLOOR]
+        v_x = conscious[c.PC_V_X] * -1
+        v_y = conscious[c.PC_V_Y] * -1
+        v_floor = conscious[c.PC_V_FLOOR] * -1
+        conscious.step()
+        yield
+        for _ in _read(rooms, conscious, rule_step_visuals):
+            yield
+        conscious[c.PC_X] = x
+        conscious[c.PC_Y] = y
+        conscious[c.PC_FLOOR] = floor
+        conscious[c.PC_V_X] = v_x
+        conscious[c.PC_V_Y] = v_y
+        conscious[c.PC_V_FLOOR] = v_floor
+        conscious.step()
+        conscious.step()
+        yield
+
+
+class WriteFlip(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(WriteFlip, self).__init__("w", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        x = conscious[c.PC_X]
+        y = conscious[c.PC_Y]
+        floor = conscious[c.PC_FLOOR]
+        v_x = conscious[c.PC_V_X] * -1
+        v_y = conscious[c.PC_V_Y] * -1
+        v_floor = conscious[c.PC_V_FLOOR] * -1
+        conscious.step()
+        yield
+        for character in ">1vur" + _write(rooms, conscious, rule_step_visuals):
+            rooms.write(*conscious.at(), character=character)
+            rule_step_visuals.append(conscious.at())
+            conscious.step()
+            yield
+        conscious[c.PC_X] = x
+        conscious[c.PC_Y] = y
+        conscious[c.PC_FLOOR] = floor
+        conscious[c.PC_V_X] = v_x
+        conscious[c.PC_V_Y] = v_y
+        conscious[c.PC_V_FLOOR] = v_floor
+        conscious.step()
+        conscious.step()
+        yield
+
+
+class DynamicDump(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(DynamicDump, self).__init__("d", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        v_floor = conscious[c.WORK_STACK].pop()
+        v_y = conscious[c.WORK_STACK].pop()
+        v_x = conscious[c.WORK_STACK].pop()
+        floor = conscious[c.WORK_STACK].pop()
+        y = conscious[c.WORK_STACK].pop()
+        x = conscious[c.WORK_STACK].pop()
+        item = conscious[c.WORK_STACK].pop()
+        if isinstance(item, int):
+            item = str(item)
+        elif item is None or item is StackBottom:
+            item = "n"
+        elif item is StackFrame:
+            item = "f"
+        if isinstance(x, int) and isinstance(y, int) and isinstance(floor, int):
+            if isinstance(v_x, int) and isinstance(v_y, int) and isinstance(v_floor, int):
+                for character in item:
+                    rooms.write(x, y, floor, character)
+                    x += v_x
+                    y += v_y
+                    floor += v_floor
+                    yield
+        conscious.step()
+        yield
+
+
+class DoubleDuplicate(Rule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(DoubleDuplicate, self).__init__("o", work_space)
+
+    def __call__(self,
+                 portal: 'backrooms.portal.Portal',
+                 rooms: Rooms,
+                 conscious: c.Conscious,
+                 rule_step_visuals: List[Tuple[int, int, int]]) -> Generator[None, None, None]:
+        """
+        info: Runs a rule.
+        :param portal: Portal
+        :param rooms: Rooms
+        :param conscious: Conscious
+        :param rule_step_visuals: List[Tuple[int, int, int]]
+        :return: Generator[None, None, None]
+        """
+        item_2 = conscious[c.WORK_STACK].pop()
+        item_1 = conscious[c.WORK_STACK].pop()
+        for _ in range(2):
+            conscious[c.WORK_STACK].push(item_1)
+            conscious[c.WORK_STACK].push(item_2)
+        conscious.step()
+        yield
+
+
+class UncommonModule(RuleModule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(UncommonModule, self).__init__("u",
+                                             work_space,
+                                             (ReadFlip,
+                                              WriteFlip,
+                                              DynamicDump,
+                                              DoubleDuplicate))
+
+
+class FrameModule(RuleModule):
+    def __init__(self,
+                 work_space: WorkSpace):
+        super(FrameModule, self).__init__("r",
+                                          work_space,
+                                          ())
 
 
 RULES = (Halt,
@@ -1343,6 +1988,7 @@ RULES = (Halt,
          Left,
          Up,
          Down,
+         DownUpper,
          Upper,
          Lower,
          Clear,
@@ -1356,7 +2002,9 @@ RULES = (Halt,
          IsStackFrame,
          IsStackBottom,
          Pop,
+         PopFrame,
          Duplicate,
+         Switch,
          Worker,
          Keep,
          Store,
@@ -1368,4 +2016,6 @@ RULES = (Halt,
          Y,
          Floor,
          ThreadModule,
-         HallwayModule)
+         HallwayModule,
+         LevelModule,
+         UncommonModule)
